/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/baseframework/view/baseview', [
    'io.ox/core/notifications',
    'io.ox/office/tk/utils',
    'io.ox/office/tk/keycodes',
    'io.ox/office/tk/forms',
    'io.ox/office/tk/dialogs',
    'io.ox/office/tk/object/triggerobject',
    'io.ox/office/baseframework/utils/errormessages',
    'io.ox/office/baseframework/app/appobjectmixin',
    'io.ox/office/baseframework/view/viewobjectmixin',
    'gettext!io.ox/office/baseframework/main',
    'less!io.ox/office/baseframework/view/basestyle'
], function (Notifications, Utils, KeyCodes, Forms, Dialogs, TriggerObject, ErrorMessages, AppObjectMixin, ViewObjectMixin, gt) {

    'use strict';

    // global functions =======================================================

    /**
     * Converts the passed number or object to a complete margin descriptor
     * with the properties 'left', 'right', 'top', and 'bottom'.
     *
     * @param {Number|Object} margin
     *  The margins, as number (for all margins) or as object with the optional
     *  properties 'left', 'right', 'top', and 'bottom'. Missing properties
     *  default to the value zero.
     *
     * @returns {Object}
     *  The margins (in pixels), in the properties 'left', 'right', 'top', and
     *  'bottom'.
     */
    function getMarginFromValue(margin) {

        if (_.isObject(margin)) {
            return _.extend({ left: 0, right: 0, top: 0, bottom: 0 }, margin);
        }

        if (!_.isNumber(margin)) {
            margin = 0;
        }
        return { left: margin, right: margin, top: margin, bottom: margin };
    }

    // class BaseView =========================================================

    /**
     * Base class for the view instance of an office application. Creates the
     * application window, and provides functionality to create and control the
     * top, bottom, and side pane elements.
     *
     * Triggers the following events:
     * - 'refresh:layout': After this view instance has refreshed the layout of
     *      all registered view panes. This event will be triggered after
     *      inserting new view panes into this view, or content nodes into the
     *      application pane, after showing/hiding view panes, while and after
     *      the browser window is resized.
     *
     * @constructor
     *
     * @extends TriggerObject
     * @extends AppObjectMixin
     * @extends ViewObjectMixin
     *
     * @param {BaseApplication} app
     *  The application containing this view instance.
     *
     * @param {BaseModel} docModel
     *  The document model created by the passed application.
     *
     * @param {Object} [initOptions]
     *  Optional parameters to control the appearance of the view:
     *  @param {String} [initOptions.classes]
     *      Additional CSS classes to be added to the root node of the
     *      application window.
     *  @param {Function} [initOptions.initHandler]
     *      A callback handler function called to initialize this view instance
     *      after construction. May return a promise to be able to run
     *      asynchronous code during initialization.
     *  @param {Function} [initOptions.initGuiHandler]
     *      A callback handler function called to initialize the graphical
     *      elements of this view instance, after the document has been
     *      imported. May return a promise to be able to run asynchronous code
     *      during initialization.
     *  @param {Function} [initOptions.grabFocusHandler]
     *      A callback function that has to implement moving the browser focus
     *      somewhere into the application pane. Used in the public method
     *      BaseView.grabFocus(). If omitted, the application pane itself will
     *      be focused.
     *  @param {Boolean} [initOptions.contentFocusable=false]
     *      If set to true, the container node for the document contents will
     *      be focusable and will be registered for global focus traversal with
     *      the F6 key.
     *  @param {Boolean} [initOptions.contentScrollable=false]
     *      If set to true, the container node for the document contents will
     *      be scrollable. By default, the size of the container node is locked
     *      and synchronized with the size of the application pane (with regard
     *      to content margins, see the option 'contentMargin' for details).
     *  @param {Number|Object} [initOptions.contentMargin=0]
     *      The margins between the fixed application pane and the embedded
     *      application container node, in pixels. If set to a number, all
     *      margins will be set to the specified value. Otherwise, an object
     *      with the optional properties 'left', 'right', 'top', and 'bottom'
     *      for specific margins for each border. Missing properties default to
     *      the value zero. The content margin can also be modified at runtime
     *      with the method BaseView.setContentMargin().
     */
    function BaseView(app, docModel, initOptions) {

        // self reference
        var self = this;

        // moves the browser focus into a node of the application pane
        var grabFocusHandler = Utils.getFunctionOption(initOptions, 'grabFocusHandler');

        // the application window instance (ox.ui.Window)
        var appWindow = app.getWindow();

        // the DOM node of the application window body
        var winBodyNode = app.getWindowNode();

        // root node of the application pane (remaining space for document contents)
        var appPaneNode = null;

        // root container node for invisible document contents
        var hiddenRootNode = $('<div class="app-hidden-root">');

        // root container node for visible document contents (may be scrollable)
        var contentRootNode = $('<div class="app-content-root">');

        // container node for application contents
        var appContentNode = $('<div class="app-content">').appendTo(contentRootNode);

        // the temporary container for all nodes while application is hidden
        var tempNode = $('<div>');

        // all registered view panes, in insertion order
        var viewPanes = [];

        // whether refreshing the pane layout is currently locked
        var layoutLocks = 0;

        // whether refreshing the pane layout is currently running
        var layoutRunning = false;

        // whether refreshing the pane layout was requested while it was locked
        var layoutPending = false;

        // whether the content root node is focusable by itself
        var contentFocusable = Utils.getBooleanOption(initOptions, 'contentFocusable', false);

        // whether the application is hidden explicitly
        var viewHidden = false;

        // cached yell options, shown when application becomes visible
        var pendingYellOptions = null;

        // the timer waiting to fade in the blocker element in busy mode
        var blockerFadeTimer = null;

        // the timer waiting to show a warning message during busy mode
        var blockerWarningTimer = null;

        // base constructors --------------------------------------------------

        TriggerObject.call(this, app);
        AppObjectMixin.call(this, app);

        // private methods ----------------------------------------------------

        function silentRefreshPaneLayout() {

            // prevent recursive processing
            if (layoutRunning) { return; }
            layoutRunning = true;

            // current offsets representing available space in the application window
            var offsets = { top: 0, bottom: 0, left: 0, right: 0 };

            // update all registered view panes
            viewPanes.forEach(function (pane) {

                var paneNode = pane.getNode(),
                    visible = pane.isVisible(),
                    position = pane.getPosition(),
                    vertical = pane.isVerticalPosition(),
                    leading = pane.isLeadingPosition(),
                    sizeFunc = _.bind(vertical ? paneNode.outerHeight : paneNode.outerWidth, paneNode),
                    paneOffsets = _.clone(offsets);

                // remove the position attribute at the opposite border of the pane position
                paneOffsets[vertical ? (leading ? 'bottom' : 'top') : (leading ? 'right' : 'left')] = '';

                // set pane position, and let pane update its internal size
                // (this may even change the pane size again due to internal refresh)
                paneNode.css(paneOffsets);
                if (visible) {
                    pane.layout();
                    offsets[position] += sizeFunc();
                }
            });

            // update the application pane node (remaining space without panes)
            appPaneNode.css(offsets);

            // leave recursion guard
            layoutRunning = false;
        }

        /**
         * Adjusts the positions of all view pane nodes.
         */
        function refreshPaneLayout() {

            // do nothing if the window is hidden (but refresh if import runs (fast loading))
            if (!app || app.isInQuit() || !self || !self.isVisible()) {
                return;
            }

            // do nothing if refreshing is currently locked, but set the pending flag
            if (layoutLocks > 0) {
                layoutPending = true;
                return;
            }

            // refresh all panes, and notify listeners
            silentRefreshPaneLayout();
            self.trigger('refresh:layout');
        }

        /**
         * Refreshes the view layout after a view pane has been modified.
         */
        var paneResizeHandler = (function () {

            // whether to really refresh the layout in the deferred callback
            var refreshLayout = false;

            function registerRefresh() {
                // do nothing when view panes are changing during layout updates
                if (!layoutRunning) { refreshLayout = true; }
            }

            function executeRefresh() {
                if (refreshLayout) {
                    refreshLayout = false;
                    refreshPaneLayout();
                }
            }

            return self.createDebouncedMethod('BaseView.paneResizeHandler', registerRefresh, executeRefresh);
        }());

        /**
         * Shows the specified notification message. If the message is of type
         * 'error', the message will be stored internally and automatically
         * shown again, after the application has been hidden and shown.
         *
         * @param {Object} yellOptions
         *  The notification message data. See method BaseView.yell() for
         *  details.
         */
        function showNotification(yellOptions) {

            if (yellOptions.type === 'error') {
                // Bug 28554: no auto-close for error messages
                yellOptions.duration = -1;
                // remember error message, show again after switching applications
                pendingYellOptions = yellOptions;
            } else {
                pendingYellOptions = null;
            }

            // create and show the notification DOM node
            var yellNode = Notifications.yell(yellOptions).addClass('io-ox-office-alert');

            // wait for deferred initialization of the DOM node
            self.listenTo(yellNode, 'notification:appear', function () {

                // the message node as target for additional contents
                var messageNode = yellNode.find('.message');

                // options for a static hyperlink shown below the message text
                var linkOptions = Utils.getObjectOption(yellOptions, 'hyperlink', null);
                // the URL of the static hyperlink
                var linkUrl = Utils.getStringOption(linkOptions, 'url', null);
                // the display text for the static hyperlink
                var linkLabel = Utils.getStringOption(linkOptions, 'label', null);

                // append the mark-up for a static hyperlink to the alert message
                if (linkUrl) {
                    linkLabel = Utils.escapeHTML(linkLabel || _.noI18n(linkUrl));
                    messageNode.append('<div><a href="' + linkUrl + '" target="_blank" rel="noopener" tabindex="0">' + linkLabel + '</a></div>');
                }

                // extended message text for better support
                var extendedMessage = Utils.getStringOption(yellOptions, 'extendedMessage');
                if (extendedMessage) {
                    messageNode.append($('<div class="small-text"><span>' + Utils.escapeHTML(extendedMessage) + '</span></div>'));
                }

                var
                    closeActionOptions = Utils.getObjectOption(yellOptions, 'closeAction'),
                    isQuitAppOnClose = Utils.getBooleanOption(closeActionOptions, 'isQuitApp', false),
                    closeHandler = Utils.getFunctionOption(closeActionOptions, 'handler', null);

                if (isQuitAppOnClose) {
                    if (closeHandler !== null) {

                        closeHandler = (function (app, docModel, callback) {
                            return function () {

                                callback(app, docModel);
                                app.quit();
                            };
                        }(app, docModel, closeHandler));

                    } else {

                        closeHandler = (function (app) {
                            return function () {

                                app.quit();
                            };
                        }(app));
                    }
                } else if (closeHandler !== null) {

                    closeHandler = (function (app, docModel, callback) {
                        return function () {

                            callback(app, docModel);
                        };
                    }(app, docModel, closeHandler));
                }
                if (closeHandler !== null) {

                    self.listenTo(yellNode, 'notification:removed', closeHandler);
                }

                // options for an action link shown below the message text
                var actionOptions = Utils.getObjectOption(yellOptions, 'action');
                // the key of the controller item for an action link
                var actionKey = Utils.getStringOption(actionOptions, 'itemKey');
                // do not insert an action link, if the controller item is missing or disabled
                if (!actionKey || !self.isControllerItemEnabled(actionKey)) { return; }

                // the icon class for the link
                var actionIcon = Utils.getStringOption(actionOptions, 'icon');
                // the label text for the link
                var actionLabel = Utils.getStringOption(actionOptions, 'label');
                // the anchor element to be inserted into the alert node
                var anchorNode = $('<a href="#">');

                // insert the icon and label to the anchor node
                if (actionIcon) { anchorNode.append(Forms.createIconMarkup(actionIcon)); }
                if (actionLabel) { anchorNode.append(Forms.createSpanMarkup(actionLabel)); }

                // execute the controller item on click (unbind the listener when quitting the application)
                self.listenTo(anchorNode, 'click tap', function (event) {
                    event.preventDefault();
                    Notifications.yell.close();
                    self.executeControllerItem(actionKey);
                });

                // insert the link into the alert node
                messageNode.append($('<div>').append(anchorNode));
            });

            // return focus to application after the notification closes
            self.listenTo(yellNode, 'notification:removed', function (event) {
                // bug 33735: only if focus is inside the alert box (e.g. after mouse click)
                if (Forms.containsFocus(event.currentTarget)) {
                    self.executeDelayed(function () { self.grabFocus(); }, 'BaseView.showNotification');
                }
            });
        }

        /**
         * Sets an invalid (negative) 'tabindex' attribute to all elements in
         * the application window body that contain a 'tabindex' attribute.
         */
        function invalidateTabIndexes() {
            winBodyNode.find('[tabindex="1"]').attr('tabindex', -42);
        }

        /**
         * Restores all 'tabindex' attributes in the application window body
         * that have been invalidated with the method invalidateTabIndexes().
         */
        function restoreTabIndexes() {
            winBodyNode.find('[tabindex="-42"]').attr('tabindex', 1);
        }

        /**
         * Updates the view after the application becomes active/visible.
         */
        function windowShowHandler() {

            // do not show the window contents if view is still hidden explicitly
            if (!self.isVisible() || app.isInQuit()) { return; }

            // bug 29348: show pending notification also, if tempNode is empty
            if (pendingYellOptions && self.isImportFinished()) {
                showNotification(pendingYellOptions);
            }

            // move all application nodes from temporary storage into view
            if (tempNode.children().length > 0) {
                winBodyNode.append(tempNode.children());
                // restore invalidated tabindex attribute for browser focus tab chain
                if (!self.isBusy()) {
                    restoreTabIndexes();
                    // bug #49359 - due to the detached DOM calling idle() didn't remove the disabled attributes from input and textarea elements, so call it again after the DOM is attached.
                    self.leaveBusy();
                }
            }

            // do not update GUI and grab focus while document is still being imported
            if (self.isImportSucceeded()) {
                self.executeDelayed(function () {
                    refreshPaneLayout();
                    app.getController().update();
                    self.grabFocus({ afterImport: true });
                }, 'BaseView.windowShowHandler');
            }
        }

        /**
         * Updates the view after the application becomes inactive/hidden.
         */
        function windowHideHandler() {
            // move all application nodes from view to temporary storage
            if (tempNode.children().length === 0) {
                // remove all nodes from browser focus tab chain as long as they are hidden
                invalidateTabIndexes();
                tempNode.append(winBodyNode.children());
            }
        }

        // protected methods --------------------------------------------------

        /**
         * Prepares the view for importing the document. Sets the application
         * to busy state, and invokes the initialization handler passed to the
         * constructor.
         *
         * @internal
         *  Invoked from the import process of the application. DO NOT call
         *  from external code.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the initialization handler has
         *  finished.
         */
        this.hideBeforeImport = _.once(function () {

            // set window to busy state
            app.getWindow().busy();
            self.hide();

            // invoke the initialization handler passed to the constructor
            var initHandler = Utils.getFunctionOption(initOptions, 'initHandler', null);
            var result = initHandler ? initHandler.call(self) : null;

            // return a promise
            return self.convertToPromise(result);
        });

        /**
         * Initializes the view after importing the document. Invokes the GUI
         * initialization handler passed to the constructor, and sets the
         * application to idle state.
         *
         * @internal
         *  Invoked from the import process of the application. DO NOT call
         *  from external code.
         *
         * @param {Boolean} [globalFailure=false]
         *  If set to true, the application is in an internal error state, and
         *  the view will not be initialized further.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the GUI initialization handler
         *  has finished.
         */
        this.showAfterImport = _.once(function (globalFailure) {

            // show the application window
            self.show();

            // invoke the GUI initialization handler passed to the constructor
            var initGuiHandler = globalFailure ? null : Utils.getFunctionOption(initOptions, 'initGuiHandler', null);
            var result = initGuiHandler ? initGuiHandler.call(self) : null;

            // wait for the callback, immediately refresh all view panes to reduce flicker effects
            return self.convertToPromise(result).always(function () {
                silentRefreshPaneLayout();
                self.leaveBusy();
            });
        });

        // public methods -----------------------------------------------------

        /**
         * Returns the application instance that owns this document view.
         *
         * @returns {BaseApplication}
         *  The application instance that owns this document view.
         */
        this.getApp = function () {
            return app;
        };

        /**
         * Returns the model of the document represented by this document view.
         *
         * @returns {BaseModel}
         *  The model of the document represented by this document view.
         */
        this.getDocModel = function () {
            return docModel;
        };

        /**
         * Returns whether the browser focus is located inside the application
         * pane.
         *
         * @returns {Boolean}
         *  Whether the browser focus is located inside the application pane.
         */
        this.hasAppFocus = function () {
            return Forms.hasOrContainsFocus(this.getAppPaneNode());
        };

        /**
         * Moves the browser focus into the application pane. Calls the
         * handler function passed to the constructor of this instance.
         *
         * @param {Object} [options]
         *  Optional parameters that will be passed to the grab focus handler
         *  function.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.grabFocus = function (options) {
            if (!app.isInQuit() && this.isVisible()) {
                var dialogNode = Dialogs.getActiveModalDialogNode();
                if (dialogNode.length > 0) {
                    Forms.grabFocus(dialogNode);
                } else if (this.isBusy()) {
                    Utils.setFocus(appWindow.nodes.blocker);
                } else if (_.isFunction(grabFocusHandler)) {
                    grabFocusHandler.call(this, options);
                } else if (contentFocusable) {
                    Utils.setFocus(contentRootNode);
                } else {
                    Forms.grabFocus(this.getAppPaneNode());
                }
            }
            return this;
        };

        /**
         * Returns whether the contents of the view are currently visible.
         *
         * @returns {Boolean}
         *  Whether the application is active, and the view has not been hidden
         *  manually.
         */
        this.isVisible = function () {
            return !viewHidden && app.isActive();
        };

        /**
         * Hides all contents of the application window and moves them to the
         * internal temporary DOM storage node.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.hide = function () {
            viewHidden = true;
            windowHideHandler();
            return this;
        };

        /**
         * Shows all contents of the application window, if the application
         * itself is currently active. Otherwise, the contents will be shown
         * when the application becomes visible.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.show = function () {
            viewHidden = false;
            windowShowHandler();
            return this;
        };

        /**
         * Sets the application into the busy state by displaying a window
         * blocker element covering the entire GUI of the application. The
         * contents of the header and footer in the blocker element are
         * cleared, and the passed initialization callback function may insert
         * new contents into these elements.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {Function} [options.initHandler]
         *      A function that can fill custom contents into the header and
         *      footer of the window blocker element. Receives the following
         *      parameters:
         *      (1) {jQuery} headerNode
         *          The header element above the centered progress bar.
         *      (2) {jQuery} footerNode
         *          The footer element below the centered progress bar.
         *      (3) {jQuery} blockerNode
         *          The entire window blocker element (containing the header,
         *          footer, and progress bar elements).
         *      Will be called in the context of this view instance.
         *  @param {Function} [options.cancelHandler]
         *      If specified, a 'Cancel' button will be shown after a short
         *      delay. Pressing that button, or pressing the ESCAPE key, will
         *      execute this callback function, and will leave the busy mode
         *      afterwards. Will be called in the context of this view
         *      instance.
         *  @param {Number} [options.delay=500]
         *      The delay time before the busy blocker element becomes visible
         *      (by fading to a half-transparent fill color). Note that the
         *      busy blocker element exists and blocks everything right after
         *      calling this method, the delay time just specifies when it
         *      becomes visible.
         *  @param {Boolean} [options.immediate=false]
         *      If set to true, the half-transparent busy blocker element will
         *      be shown immediately. Otherwise, the element will be set to
         *      transparent initially, and will fade in after the specified
         *      delay (see option 'delay').
         *  @param {Boolean} [options.showFileName=false]
         *      If set to true, the file name will be shown in the top-left
         *      corner of the blocker element.
         *  @param {String} [options.warningLabel]
         *      If specified, an alert box containing the passed text will be
         *      shown after a specific delay (see option 'warningDelay').
         *      Intended to be used to show a message for long-running actions.
         *  @param {Number} [options.warningDelay=3000]
         *      The delay time before the warning alert becomes visible. If the
         *      blocker element will be faded in delayed by itself, the warning
         *      delay will start after that delay time.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.enterBusy = function (options) {

            // for safety against repeated calls: stops pending animations etc.
            this.leaveBusy();

            // enter busy state, and extend the blocker element
            appWindow.busy(null, null, function (blockerNode, headerNode, footerNode) {

                // the initialization handler
                var initHandler = Utils.getFunctionOption(options, 'initHandler');
                // the cancel handler
                var cancelHandler = Utils.getFunctionOption(options, 'cancelHandler');
                // the delay time before the blocker element becomes visible
                var delay = Utils.getIntegerOption(options, 'delay', 500, 0);
                // the label text for the warning alert for long-running actions
                var warningLabel = Utils.getStringOption(options, 'warningLabel', '');
                // the delay time before the warning alert box becomes visible
                var warningDelay = Utils.getIntegerOption(options, 'warningDelay', 3000, 0);
                // the DOM element containing the warning alert box, or the single cancel button
                var warningNode = $('<div>').hide();
                // the cancel button element
                var cancelButton = null;

                // focus handler that prevents focusing document contents during busy mode
                function changeFocusHandler(event, focusNode) {
                    if (focusNode && Utils.containsNode(winBodyNode, focusNode) && !Utils.containsNode(blockerNode, focusNode, { allowEqual: true })) {
                        Utils.setFocus(((cancelButton && cancelButton.is(':visible')) ? cancelButton : blockerNode));
                    }
                }

                // keyboard event handler for busy mode (ESCAPE key)
                function keyDownHandler(event) {
                    if (cancelHandler && (event.keyCode === KeyCodes.ESCAPE)) {
                        cancelHandler.call(self);
                        return false;
                    }
                }

                // unregister all event listeners
                function unregisterListeners() {
                    self.stopListeningTo(Utils, 'change:focus', changeFocusHandler);
                    if (self.isVisible()) { restoreTabIndexes(); }
                    self.stopListeningTo(blockerNode, 'keydown', keyDownHandler);
                    self.stopListeningTo(appWindow, 'idle', unregisterListeners);
                }

                // bug 38770: grab focus to prevent modifications of document
                Utils.setFocus(blockerNode.attr('tabindex', 1));
                invalidateTabIndexes();
                self.listenTo(Utils, 'change:focus', changeFocusHandler);

                // bug 41397: destroying browser selection helps to prevent document modifications
                Utils.clearBrowserSelection();

                // register a keyboard handler for the ESCAPE key
                self.listenTo(blockerNode, 'keydown', keyDownHandler);

                // unregister all event handlers when leaving busy mode
                self.listenTo(appWindow, 'idle', unregisterListeners);

                // execute initialization handler
                if (_.isFunction(initHandler)) {
                    initHandler.call(self, headerNode, footerNode, blockerNode);
                }

                // append the container node for the warning alert box and/or cancel button
                footerNode.append(warningNode);

                // fade in blocker node after a delay
                if (!Utils.getBooleanOption(options, 'immediate', false)) {
                    blockerNode.css('opacity', 0);
                    blockerFadeTimer = self.repeatDelayed(function (index) {
                        blockerNode.css('opacity', (index + 1) / 10);
                    }, 'BaseView.blockerFadeTimer', { delay: delay, repeatDelay: 50, cycles: 10 });
                    // defer showing the warning box until blocker is visible
                    warningDelay += delay + 500;
                }

                // initialize a cancel button, if a callback handler has been passed
                if (cancelHandler) {
                    cancelButton = $.button({ label: gt('Cancel') })
                        .addClass('btn-warning')
                        .on('click', function () { cancelHandler.call(self); });
                }

                // create a warning alert box if specified, insert cancel button into the alert box
                if (warningLabel.length > 0) {
                    var alertNode = $('<div class="alert alert-warning"><div>' + Utils.escapeHTML(warningLabel) + '</div></div>');
                    if (cancelButton) { alertNode.append($('<div>').append(cancelButton)); }
                    warningNode.append(alertNode);
                } else if (cancelButton) {
                    warningNode.append(cancelButton);
                }

                // show warning alert box after a delay
                if (warningNode.children().length > 0) {
                    blockerWarningTimer = self.executeDelayed(function () {
                        warningNode.show();
                        if (cancelButton) {
                            Utils.setFocus(cancelButton);
                            blockerNode.removeAttr('tabindex');
                        }
                    }, 'BaseView.blockerWarningTimer', warningDelay);
                }
            }, options);

            return this;
        };

        /**
         * Returns whether this application is currently in busy state.
         *
         * @returns {Boolean}
         *  Whether this application is currently in busy state.
         */
        this.isBusy = function () {
            return appWindow.isBusy();
        };

        /**
         * Updates the progress bar of the busy blocker element.
         *
         * @param {Number} progress
         *  The progress value (floating-point number between 0 and 1).
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.updateBusyProgress = (function () {

            // the current progress value to be shown at the progress bar
            var currProgress = 0;

            function registerProgress(progress) {
                currProgress = progress;
                return self;
            }

            function showProgress() {
                if (self.isBusy()) {
                    appWindow.busy(currProgress);
                }
            }

            return self.createDebouncedMethod('BaseView.updateBusyProgress', registerProgress, showProgress, { delay: 500, maxDelay: 500 });
        }());

        /**
         * Leaves the busy state of the application. Hides the window blocker
         * element covering the entire GUI of the application.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.leaveBusy = function () {

            // stop timer that fades in the blocker element delayed
            if (blockerFadeTimer) {
                blockerFadeTimer.abort();
                blockerFadeTimer = null;
            }

            // stop timer that wants to show a warning alert in the blocker element
            if (blockerWarningTimer) {
                blockerWarningTimer.abort();
                blockerWarningTimer = null;
            }

            appWindow.idle();
            return this;
        };

        /**
         * Shows an alert banner, if the application is currently visible.
         * Otherwise, the last alert banner will be cached until the
         * application becomes visible again.
         *
         * @param {Object} yellOptions
         *  The settings for the alert banner:
         *  @param {String} [yellOptions.type='info']
         *      The type of the alert banner. Supported types are 'success',
         *      'info', 'warning', and 'error'.
         *  @param {String} [yellOptions.headline]
         *      An optional headline shown above the message text.
         *  @param {String} yellOptions.message
         *      The message text shown in the alert banner.
         *  @param {Number} [yellOptions.duration]
         *      The time to show the alert banner, in milliseconds; or -1 to
         *      show a permanent alert. Default duration is dependent on the
         *      alert type: permanent for error alerts, and specific durations
         *      defined by the core toolkit for all other types.
         *  @param {Object} [yellOptions.hyperlink]
         *      An arbitrary static hyperlink that will be shown below the
         *      message text, with the following properties:
         *      @param {String} yellOptions.hyperlink.url
         *          The URL for the hyperlink.
         *      @param {String} [yellOptions.hyperlink.label]
         *          The display text shown as hyperlink. If omitted, the URL
         *          will be displayed.
         *  @param {Object} [yellOptions.action]
         *      A descriptor for an action link shown below the message text.
         *      The link will be bound to an arbitrary item of the application
         *      controller. If specified, must be an object with the following
         *      properties:
         *      @param {String} yellOptions.action.itemKey
         *          The key of a controller item that will be bound to the
         *          link. The link will be shown if the controller item is
         *          enabled, and upon clicking the link, the controller item
         *          will be executed.
         *      @param {String} [yellOptions.action.icon]
         *          The class name of an icon shown before the link label. MUST
         *          be an icon from FontAwesome, internal OX Documents bitmap
         *          icons are NOT supported.
         *      @param {String} yellOptions.action.label
         *          A label text for the link.
         *  @param {Object} [yellOptions.closeAction]
         *      A descriptor for how to handle an actively triggered close event of any notification popup.
         *      If specified, it should be an object with the following properties:
         *      @param {Boolean} yellOptions.closeAction.isQuitApp
         *          An optional boolean flag pointing to whether the current application should be quit to or not.
         *      @param {Function} yellOptions.closeAction.handler
         *          An optionally provided method, handled as callback. It has to provide behavior that will be
         *          executed in addition to closing the notification popup. Following arguments will be passed
         *          into such a closing action handler ... [{app}, {docModel}, {BaseView}].
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.yell = function (yellOptions) {

            // add default options
            yellOptions = _.extend({ type: 'info' }, yellOptions);

            // only show notification if the application is active, otherwise cache for later use
            if (this.isVisible()) {
                showNotification(yellOptions);
            } else {
                pendingYellOptions = yellOptions;
            }
            return this;
        };

        /**
         * Hides the alert banner currently shown, and clears the cached
         * settings for an alert banner to be shown in the future when this
         * application becomes visible.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.hideYell = function () {
            Notifications.yell.close();
            pendingYellOptions = null;
            return this;
        };

        /**
         * A shortcut to receive the enabled state of the specified controller
         * item that directly invokes the method BaseController.isItemEnabled()
         * of the application controller owning this view instance.
         *
         * @param {String} key
         *  The key of the item.
         *
         * @returns {Boolean}
         *  Whether the specified item exists and is enabled.
         */
        this.isControllerItemEnabled = function (key) {
            return app.getController().isItemEnabled(key);
        };

        /**
         * A shortcut to receive the value of the specified controller item
         * that directly invokes the method BaseController.getItemValue() of
         * the application controller owning this view instance.
         *
         * @param {String} key
         *  The key of the item.
         *
         * @returns {Any}
         *  The current value of the item; or undefined, if the item does not
         *  exist.
         */
        this.getControllerItemValue = function (key) {
            return app.getController().getItemValue(key);
        };

        /**
         * A shortcut to execute the specified controller item that directly
         * invokes the method BaseController.executeItem() of the application
         * controller owning this view instance.
         *
         * @param {String} key
         *  The key of the controller item to be changed.
         *
         * @param {Any} [value]
         *  The new value of the controller item.
         *
         * @param {Object} [options]
         *  Optional parameters passed to the controller.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved or rejected according to the result
         *  of the item setter function.
         */
        this.executeControllerItem = function (key, value, options) {
            return app.getController().executeItem(key, value, options);
        };

        /**
         * Adds the passed view pane instance into this view.
         *
         * @param {Pane} pane
         *  The view pane instance to be inserted into this view.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.addPane = function (pane) {
            // insert the new view pane
            viewPanes.push(pane);
            (viewHidden ? tempNode : winBodyNode).append(pane.getNode());

            // refresh pane layout when the pane or its contents are changed
            pane.on('pane:show pane:hide pane:resize', paneResizeHandler);
            paneResizeHandler();

            return this;
        };

        /**
         * Prevents refreshing the pane layout while the specified callback
         * function is running. After the callback function has finished, the
         * pane layout will be adjusted once by calling the method
         * BaseView.refreshPaneLayout().
         *
         * @param {Function} callback
         *  The callback function that will be executed synchronously while
         *  refreshing the pane layout is locked.
         *
         * @param {Object} [context]
         *  The context bound to the callback function.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.lockPaneLayout = function (callback, context) {
            layoutLocks += 1;
            callback.call(context);
            layoutLocks -= 1;
            if (layoutLocks === 0) {
                if (layoutPending) { refreshPaneLayout(); }
                layoutPending = false;
            }
            return this;
        };

        /**
         * Adjusts the positions of all view pane nodes.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.refreshPaneLayout = function () {
            refreshPaneLayout();
            return this;
        };

        /**
         * Returns the DOM node of the application pane (the complete inner
         * area between all existing view panes). Note that this is NOT the
         * container node where applications insert their own contents. The
         * method BaseView.insertContentNode() is intended to be used to insert
         * own contents into the application pane.
         *
         * @returns {jQuery}
         *  The DOM node of the application pane.
         */
        this.getAppPaneNode = function () {
            return appPaneNode;
        };

        /**
         * Detaches the document contents in the application pane from the DOM.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.detachAppPane = function () {
            appContentNode.detach();
            return this;
        };

        /**
         * Attaches the document contents in the application pane to the DOM.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.attachAppPane = function () {
            contentRootNode.append(appContentNode);
            return this;
        };

        /**
         * Returns the root container node for the application contents. Note
         * that this is NOT the direct parent node where applications insert
         * their own contents, but the (optionally scrollable) root node
         * containing the target container node for document contents. The
         * method BaseView.insertContentNode() is intended to be used to insert
         * own contents into the application pane.
         *
         * @returns {jQuery}
         *  The DOM root node for visible document contents.
         */
        this.getContentRootNode = function () {
            return contentRootNode;
        };

        /**
         * Returns the container node for application contents
         *
         * @returns {jQuery}
         *  The DOM root node for application contents
         */
        this.getAppContentNode = function () {
            return appContentNode;
        };

        /**
         * Returns the current margins between the fixed application pane and
         * the embedded application container node.
         *
         * @returns {Object}
         *  The margins between the fixed application pane and the embedded
         *  application container node (in pixels), in the properties 'left',
         *  'right', 'top', and 'bottom'.
         */
        this.getContentMargin = function () {
            return {
                left: Utils.getElementCssLength(appContentNode, 'margin-left'),
                right: Utils.getElementCssLength(appContentNode, 'margin-right'),
                top: Utils.getElementCssLength(appContentNode, 'margin-top'),
                bottom: Utils.getElementCssLength(appContentNode, 'margin-bottom')
            };
        };

        /**
         * Changes the margin between the fixed application pane and the
         * embedded application container node.
         *
         * @param {Number|Object} margin
         *  The margins between the fixed application pane and the embedded
         *  application container node, in pixels. If set to a number, all
         *  margins will be set to the specified value. Otherwise, an object
         *  with the optional properties 'left', 'right', 'top', and 'bottom'
         *  for specific margins for each border. Missing properties default to
         *  the value zero.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.setContentMargin = function (margin) {
            margin = getMarginFromValue(margin);
            appContentNode.css('margin', margin.top + 'px ' + margin.right + 'px ' + margin.bottom + 'px ' + margin.left + 'px');
            return this;
        };

        /**
         * Inserts new DOM nodes into the container node of the application
         * pane.
         *
         * @param {HTMLElement|jQuery} contentNode
         *  The DOM node(s) to be inserted into the application pane. If this
         *  object is a jQuery collection, inserts all contained DOM nodes into
         *  the application pane.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.insertContentNode = function (contentNode) {
            appContentNode.append(contentNode);
            refreshPaneLayout();
            return this;
        };

        /**
         * Removes all DOM nodes from the container node of the application
         * pane.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.removeAllContentNodes = function () {
            appContentNode.empty();
            refreshPaneLayout();
            return this;
        };

        /**
         * Returns the root container node for hidden application contents.
         *
         * @returns {jQuery}
         *  The DOM root node for hidden document contents.
         */
        this.getHiddenRootNode = function () {
            return hiddenRootNode;
        };

        /**
         * Inserts new DOM nodes into the hidden container node of the
         * application pane.
         *
         * @param {HTMLElement|jQuery} nodes
         *  The DOM node(s) to be inserted into the hidden container of the
         *  application pane. If this object is a jQuery collection, inserts
         *  all contained DOM nodes into the application pane.
         *
         * @returns {BaseView}
         *  A reference to this instance.
         */
        this.insertHiddenNodes = function (nodes) {
            hiddenRootNode.append(nodes);
            return this;
        };

        // initialization -----------------------------------------------------

        // mix-in classes expecting complete view
        ViewObjectMixin.call(this, this);

        // add CSS classes to the window root node, and the temporary container node
        (function () {

            // all CSS classes (always add the application name)
            var windowNodeClasses = 'io-ox-office-main ' + app.getName().replace(/[./]/g, '-') + '-main';
            // additional CSS classes from options
            var classes = Utils.getStringOption(initOptions, 'classes');

            // add the classes passed through options
            if (classes) { windowNodeClasses += ' ' + classes; }

            // add application-specific classes to the window body node, and the temporary container node
            winBodyNode.addClass(windowNodeClasses);
            tempNode.addClass(windowNodeClasses);
        }());

        // add the temporary container for hidden state into the DOM
        Utils.insertHiddenNodes(tempNode);

        // log layout events
        Utils.LayoutLogger.logEvent(this, 'refresh:layout');

        // create the application pane, and insert the container nodes
        appPaneNode = $('<div class="app-pane">');
        appPaneNode.append(hiddenRootNode, contentRootNode);
        winBodyNode.append(appPaneNode);

        // keep application in DOM while application is hidden, applications
        // may want to access element geometry in background tasks
        this.listenTo(appWindow, 'show', windowShowHandler);
        this.listenTo(appWindow, 'hide', windowHideHandler);

        // listen to browser window resize events when the application window is visible
        app.registerWindowResizeHandler(refreshPaneLayout);

        // close all app menus on orientationchange on compact devices
        // .grabFocus() has an isVisible check so it's not necessary here
        if (Utils.COMPACT_DEVICE) {
            app.registerGlobalEventHandler(window, 'orientationchange', function () {
                self.grabFocus();
            });
        }

        // initialize content node from passed options
        contentRootNode.toggleClass('scrolling', Utils.getBooleanOption(initOptions, 'contentScrollable', false));
        this.setContentMargin(Utils.getOption(initOptions, 'contentMargin', 0));

        // make the content root node focusable for global navigation with F6 key
        if (contentFocusable) {
            contentRootNode.addClass('f6-target').attr('tabindex', 1);
        }

        // update view and pane layout after import
        this.waitForImport(windowShowHandler);

        // show notification alerts when import fails
        this.waitForImportFailure(function () {

            var error = app.getErrorState();
            var fileDesc = app.getFileDescriptor();
            var messageData = ErrorMessages.getMessageData(error, { originalFileId: fileDesc ? fileDesc.id : undefined });

            if (_.isObject(messageData)) {
                app.extendMessageData(messageData);
                self.yell(messageData);
            }
        });

        // destroy all class members on destruction
        this.registerDestructor(function () {

            // destroy all image nodes (release the browser-internal bitmap data)
            app.destroyImageNodes(winBodyNode);
            app.destroyImageNodes(tempNode);

            _.invoke(viewPanes, 'destroy');

            tempNode.remove();
            appPaneNode.remove();

            self = app = docModel = initOptions = null;
            grabFocusHandler = appWindow = winBodyNode = appPaneNode = null;
            hiddenRootNode = contentRootNode = appContentNode = tempNode = null;
            viewPanes = blockerFadeTimer = blockerWarningTimer = null;
        });

    } // class BaseView

    // exports ================================================================

    // derive this class from class TriggerObject
    return TriggerObject.extend({ constructor: BaseView });

});
