/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/tk/render/canvas', [
    'io.ox/office/tk/utils',
    'io.ox/office/tk/object/baseobject',
    'io.ox/office/tk/render/colorutils',
    'io.ox/office/tk/render/path',
    'io.ox/office/tk/render/rectangle'
], function (Utils, BaseObject, ColorUtils, Path, Rectangle) {

    'use strict';

    // convenience shortcuts
    var isFunction = _.isFunction;
    var isObject = _.isObject;
    var isArray = _.isArray;
    var RGBModel = ColorUtils.RGBModel;

    // color byte array for full transparence
    var TRANSPARENT = [0, 0, 0, 0];

    // temporary canvas element used for content relocation
    var relocateCanvas = null;

    // private static functions ===============================================

    /**
     * Returns whether the passed value is a rectangle with a valid size.
     *
     * @param {Any} rectangle
     *  Any value that will be checked.
     *
     * @returns {Boolean}
     *  Whether the passed value is a rectangle with positive width and height.
     */
    function isValidRectangle(rectangle) {
        return isObject(rectangle) && (rectangle.width > 0) && (rectangle.height > 0);
    }

    /**
     * Adds the passed color stops to the canvas gradient, and catches any
     * exceptions that may occur due to invalid input values.
     *
     * @param {CanvasGradient} gradient
     *  A canvas gradient descriptor.
     *
     * @param {Array<Object>} stops
     *  The color stops, as array of objects with the properties 'offset' and
     *  'color'.
     */
    function addGradientColorStops(gradient, stops) {
        try {
            stops.forEach(function (stop) {
                gradient.addColorStop(stop.offset, stop.color);
            });
        } catch (ex) {
            Utils.exception(ex);
        }
    }

    // class PathGenerator ====================================================

    /**
     * This class implements the rendering path API of a 2D rendering context
     * of a canvas element. It uses the native methods of the wrapped rendering
     * context, but applies a fixed global translation to all operations. This
     * is needed to workaround the limited floating-point precision provided by
     * the canvas (single-precision instead of double-precision, see bug 35653,
     * and http://stackoverflow.com/a/8874802).
     *
     * @constructor
     *
     * @param {CanvasRenderingContext2D} context
     *  The wrapped rendering context of a canvas element.
     *
     * @param {Number} x0
     *  The global translation on the X axis that will be applied to all path
     *  operations.
     *
     * @param {Number} y0
     *  The global translation on the Y axis that will be applied to all path
     *  operations.
     */
    function PathGenerator(context, x0, y0) {

        // the wrapped rendering context
        this.context = context;
        // the global X translation
        this.x0 = x0;
        // the global Y translation
        this.y0 = y0;

    } // class PathGenerator

    // path API ---------------------------------------------------------------

    PathGenerator.prototype.beginPath = function () {
        this.context.beginPath();
    };

    PathGenerator.prototype.save = function () {
        this.context.save();
    };

    PathGenerator.prototype.restore = function () {
        this.context.restore();
    };

    PathGenerator.prototype.translate = function (dx, dy) {
        this.context.translate(dx, dy);
    };

    PathGenerator.prototype.lineCap = function (value) {
        this.context.lineCap = value;
    };

    PathGenerator.prototype.rotate = function (a) {
        this.context.rotate(a);
    };

    PathGenerator.prototype.moveTo = function (x, y) {
        this.context.moveTo(x - this.x0, y - this.y0);
    };

    PathGenerator.prototype.lineTo = function (x, y) {
        this.context.lineTo(x - this.x0, y - this.y0);
    };

    PathGenerator.prototype.arcTo = function (x1, y1, x2, y2, r) {
        this.context.arcTo(x1 - this.x0, y1 - this.y0, x2 - this.x0, y2 - this.y0, r);
    };

    PathGenerator.prototype.quadraticCurveTo = function (cx, cy, x, y) {
        this.context.quadraticCurveTo(cx - this.x0, cy - this.y0, x - this.x0, y - this.y0);
    };

    PathGenerator.prototype.bezierCurveTo = function (cx1, cy1, cx2, cy2, x, y) {
        this.context.bezierCurveTo(cx1 - this.x0, cy1 - this.y0, cx2 - this.x0, cy2 - this.y0, x - this.x0, y - this.y0);
    };

    PathGenerator.prototype.arc = function (x, y, r, a1, a2, ccw) {
        this.context.arc(x - this.x0, y - this.y0, r, a1, a2, ccw);
    };

    PathGenerator.prototype.ellipse = function (x, y, rx, ry, a0, a1, a2, ccw) {
        this.context.ellipse(x - this.x0, y - this.y0, rx, ry, a0, a1, a2, ccw);
    };

    PathGenerator.prototype.closePath = function () {
        this.context.closePath();
    };

    // public methods ---------------------------------------------------------

    /**
     * Generates the passed path in the wrapped rendering context using its
     * native implementation.
     *
     * @param {Path} path
     *  The description of the rendering path to be generated.
     */
    PathGenerator.prototype.generatePath = function (path) {
        this.beginPath();
        path.ops.forEach(function (op) {
            this[op.name].apply(this, op.args);
        }, this);
    };

    // class ContextWrapper ===================================================

    /**
     * This wrapper class provides a higher-level API for a 2D rendering
     * context of a canvas element.
     *
     * @constructor
     *
     * @param {CanvasRenderingContext2D} renderContext
     *  The wrapped rendering context of a canvas element.
     *
     * @param {Number} x0
     *  The global translation on the X axis that will be applied to all path
     *  operations.
     *
     * @param {Number} y0
     *  The global translation on the Y axis that will be applied to all path
     *  operations.
     */
    function ContextWrapper(renderContext, x0, y0) {

        this._context = renderContext;
        this._x0 = x0;
        this._y0 = y0;

        // generator for native paths in the rendering context
        this._pathGenerator = new PathGenerator(renderContext, x0, y0);

        // current font settings
        this._fontStyle = {
            font: '10pt sans-serif',
            align: 'start',
            baseline: 'alphabetic'
        };

    } // class ContextWrapper

    // private methods --------------------------------------------------------

    /**
     * Sets the current line dash pattern.
     *
     * @param {Number|Array<Number>|Null} pattern
     *  The new dash pattern. See method ContextWrapper.setLineStyle() for
     *  details.
     */
    ContextWrapper.prototype._initLineDash = function (pattern) {

        // convert parameter to pattern array
        if (typeof pattern === 'number') {
            pattern = [pattern];
        } else if (!isArray(pattern)) {
            pattern = [];
        }

        // initialize line dash
        this._context.setLineDash(pattern);
    };

    // public methods ---------------------------------------------------------

    /**
     * Returns the original rendering context wrapped internally.
     *
     * @returns {CanvasRenderingContext2D}
     *  The original rendering context wrapped internally.
     */
    ContextWrapper.prototype.valueOf = function () {
        return this._context;
    };

    /**
     * Returns the HTML canvas element wrapped internally.
     *
     * @returns {HTMLCanvasElement}
     *  The HTML canvas element wrapped internally.
     */
    ContextWrapper.prototype.getCanvas = function () {
        return this._context.canvas;
    };

    /**
     * Creates a descriptor for a linear gradient, that can be used as fill
     * style for the inner area of rendering paths, or as stroke style for the
     * outline of a rendering path.
     *
     * @param {Number} x1
     *  The X coordinate of the starting point of the linear gradient.
     *
     * @param {Number} y1
     *  The Y coordinate of the starting point of the linear gradient.
     *
     * @param {Number} x2
     *  The X coordinate of the end point of the linear gradient.
     *
     * @param {Number} y2
     *  The Y coordinate of the end point of the linear gradient.
     *
     * @param {Array<Object>} stops
     *  The color stops for the gradient. Each object in the array MUST contain
     *  the following properties:
     *  - {Number} stop.offset
     *      The relative position of the color stop on the line defined by the
     *      coordinates passed to this method. MUST be a floating-point number
     *      in the interval [0;1]. The value 0 represents the starting point of
     *      the line, and the value 1 represents its end point.
     *  - {String} stop.color
     *      The CSS color ('#RRGGBB', '#RRGGBBAA', '#RGB', '#RGBA' style, or as
     *      'rgb()', 'rgba()', 'hsl()', or 'hsla()' function style, or any
     *      predefined or system color name).
     *
     * @returns {CanvasGradient}
     *  The descriptor for a linear gradient. Can be used with the methods
     *  ContextWrapper.setFillStyle(), or ContextWrapper.setLineStyle().
     */
    ContextWrapper.prototype.createLinearGradient = function (x1, y1, x2, y2, stops) {
        var gradient = this._context.createLinearGradient(x1 - this._x0, y1 - this._y0, x2 - this._x0, y2 - this._y0);
        addGradientColorStops(gradient, stops);
        return gradient;
    };

    /**
     * Creates a descriptor for a radial gradient, that can be used as fill
     * style for the inner area of rendering paths, or as stroke style for the
     * outline of a rendering path.
     *
     * @param {Number} x1
     *  The X coordinate of the start circle of the radial gradient.
     *
     * @param {Number} y1
     *  The Y coordinate of the start circle of the radial gradient.
     *
     * @param {Number} r1
     *  The radius of the start circle of the radial gradient.
     *
     * @param {Number} x2
     *  The X coordinate of the end circle of the radial gradient.
     *
     * @param {Number} y2
     *  The Y coordinate of the end circle of the radial gradient.
     *
     * @param {Number} r2
     *  The radius of the end circle of the radial gradient.
     *
     * @param {Array<Object>} stops
     *  The color stops for the gradient. Each object in the array MUST contain
     *  the following properties:
     *  - {Number} stop.offset
     *      The relative position of the color stop between the circles defined
     *      by the coordinates passed to this method. MUST be a floating-point
     *      number in the interval [0;1]. The value 0 represents the start
     *      circle, and the value 1 represents the end circle.
     *  - {String} stop.color
     *      The CSS color ('#RRGGBB', '#RRGGBBAA', '#RGB', '#RGBA' style, or as
     *      'rgb()', 'rgba()', 'hsl()', or 'hsla()' function style, or any
     *      predefined or system color name).
     *
     * @returns {CanvasGradient}
     *  The descriptor for a radial gradient. Can be used with the methods
     *  ContextWrapper.setFillStyle(), or ContextWrapper.setLineStyle().
     */
    ContextWrapper.prototype.createRadialGradient = function (x1, y1, r1, x2, y2, r2, stops) {
        var gradient = this._context.createRadialGradient(x1 - this._x0, y1 - this._y0, r1, x2 - this._x0, y2 - this._y0, r2);
        addGradientColorStops(gradient, stops);
        return gradient;
    };

    /**
     * Creates a descriptor for a fill pattern, that can be used as fill style
     * for the inner area of rendering paths, or as stroke style for the
     * outline of a rendering path.
     *
     * @param {CanvasImageSource|jQuery|Canvas} source
     *  The bitmap source object. Can be any DOM element supporting the DOM
     *  helper type CanvasImageSource (e.g. <img>, <canvas>, or <video>
     *  elements), or a jQuery collection wrapping a DOM element mentioned
     *  before, or an instance of the toolkit class Canvas itself.
     *
     * @param {String} [repeat='repeat']
     *  The repetition mode. MUST be one of 'repeat' (repeat the bitmap in both
     *  directions), 'repeat-x' (repeat the bitmap horizontally only),
     *  'repeat-y' (repeat the bitmap vertically only), or 'no-repeat' (no
     *  repetition of the bitmap).
     *
     * @returns {CanvasPattern}
     *  The descriptor for a fill pattern. Can be used with the methods
     *  ContextWrapper.setFillStyle(), or ContextWrapper.setLineStyle().
     */
    ContextWrapper.prototype.createPattern = function (source, repeat) {

        // convert canvas wrapper or jQuery objects to canvas DOM element
        if (source instanceof Canvas) {
            source = source.getNode()[0];
        } else if (source instanceof $) {
            source = source[0];
        }

        return this._context.createPattern(source, repeat || 'repeat');
    };

    /**
     * Sets the fill style for rendering the inner area of a path.
     *
     * @param {String|CanvasGradient|CanvasPattern} style
     *  The fill style for rendering the inner fill area, as string for a solid
     *  fill color (any CSS color: as '#RRGGBB', '#RRGGBBAA', '#RGB', '#RGBA'
     *  style, or as 'rgb()', 'rgba()', 'hsl()', or 'hsla()' function style, or
     *  any predefined or system color name), or as instance of CanvasGradient
     *  for linear or radial gradient styles, or as instance of CanvasPattern
     *  for pattern or bitmap fill styles.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.setFillStyle = function (style) {
        this._context.fillStyle = style;
        return this;
    };

    /**
     * Sets the stroke style, and other line properties, for rendering the
     * outline of a path.
     *
     * @param {Object} options
     *  Stroke style parameters. Omitting a property will not change the
     *  respective style setting.
     *  @param {String|CanvasGradient|CanvasPattern} [options.style]
     *      The fill style for rendering line segments of a path, as string for
     *      a solid fill color (any CSS color: as '#RRGGBB', '#RRGGBBAA',
     *      '#RGB', '#RGBA' style, or as 'rgb()', 'rgba()', 'hsl()', or
     *      'hsla()' function style, or any predefined or system color name),
     *      or as instance of CanvasGradient for linear or radial gradient
     *      styles, or as instance of CanvasPattern for pattern or bitmap fill
     *      styles.
     *  @param {Number} [options.width]
     *      The width of the line segments.
     *  @param {Number} [options.cap]
     *      The style of the line end points. Supports all values supported by
     *      the property 'lineCap' of the native rendering context.
     *  @param {Number} [options.join]
     *      The style of the line connection points. Supports all values
     *      supported by the property 'lineJoin' of the native rendering
     *      context.
     *  @param {Array<Number>|Number|Null} [options.pattern]
     *      The line dash pattern, as pairs of numbers in a flat array. Each
     *      number in the array MUST be non-negative, and the sum of all
     *      numbers (the length of the pattern) MUST be greater than zero. The
     *      first entry of each pair in this array represents the length of a
     *      visible line segment, the second entry of each pair represents the
     *      length of the gap to the next line segment. By passing multiple
     *      pairs it is possible to specify more complex dash patterns like
     *      dash-dot-dotted lines etc. If the number of elements in the array
     *      is odd, the resulting dash pattern will be the concatenation of two
     *      copies of the array. If set to a number, uses a simple pattern with
     *      equally sized segments and gaps. If set to null, the current line
     *      dash pattern will be removed, and solid lines will be rendered.
     *  @param {Number} [options.patternOffset]
     *      The initial pattern offset for every subpath in a path stroked with
     *      a dash pattern.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.setLineStyle = function (options) {

        if ('style' in options) { this._context.strokeStyle = options.style; }
        if ('width' in options) { this._context.lineWidth = options.width; }
        if ('cap' in options) { this._context.lineCap = options.cap; }
        if ('join' in options) { this._context.lineJoin = options.join; }
        if ('pattern' in options) { this._initLineDash(options.pattern); }
        if ('patternOffset' in options) { this._context.lineDashOffset = options.patternOffset; }

        return this;
    };

    /**
     * Sets the shadow style for rendering a path in the canvas.
     *
     * @param {Object} options
     *  Shadow style parameters. Omitting a property will not change the
     *  respective style setting.
     *  @param {String} [options.color]
     *      The shadow color (any CSS color: as '#RRGGBB', '#RRGGBBAA', '#RGB',
     *      '#RGBA' style, or as 'rgb()', 'rgba()', 'hsl()', or 'hsla()'
     *      function style, or any predefined or system color name).
     *  @param {Number} [options.blur]
     *      The size of the blur effect.
     *  @param {Number} [options.offsetX]
     *      The offset of the shadow relative to the path in X direction.
     *  @param {Number} [options.offsetY]
     *      The offset of the shadow relative to the path in Y direction.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.setShadowStyle = function (options) {

        if ('color' in options) { this._context.shadowColor = options.color; }
        if ('blur' in options) { this._context.shadowBlur = options.blur; }
        if ('offsetX' in options) { this._context.shadowOffsetX = options.offsetX; }
        if ('offsetY' in options) { this._context.shadowOffsetY = options.offsetY; }

        return this;
    };

    /**
     * Sets the font style, and other font properties, for rendering of texts
     * in the canvas.
     *
     * @param {Object} options
     *  Font style parameters. Omitting a property will not change the
     *  respective style setting.
     *  @param {String} [options.font]
     *      The font styles, as expected by the property 'font' of the canvas
     *      rendering context. Initial value is '10pt sans-serif'.
     *  @param {String} [options.align]
     *      The horizontal alignment (accepts all values supported by the
     *      property 'textAlign' of a canvas rendering context).
     *  @param {String} [options.baseline]
     *      The text baseline mode (accepts all values supported by the
     *      property 'textBaseline' of a canvas rendering context).
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.setFontStyle = function (options) {

        _.extend(this._fontStyle, options);
        this._context.font = this._fontStyle.font;
        this._context.textAlign = this._fontStyle.align;
        this._context.textBaseline = this._fontStyle.baseline;

        return this;
    };

    /**
     * Sets the global opaqueness for all drawing operations (fill, stroke,
     * text, and images).
     *
     * @param {Number} alpha
     *  The opaqueness. The value 1 represents full opacity, the value 0
     *  represents full transparency.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.setGlobalAlpha = function (alpha) {
        this._context.globalAlpha = alpha;
        return this;
    };

    /**
     * Sets the global composite operation for all drawing operations (fill,
     * stroke, text, and images).
     *
     * @param {String} operation
     *  The composite operation.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.setGlobalComposite = function (operation) {
        this._context.globalCompositeOperation = operation;
        return this;
    };

    /**
     * Changes the translation of the coordinate system.
     *
     * @param {Number} dx
     *  The horizontal translation.
     *
     * @param {Number} dy
     *  The vertical translation.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.translate = function (dx, dy) {
        this._context.translate(dx, dy);
        return this;
    };

    /**
     * Changes the scaling factors of the coordinate system.
     *
     * @param {Number} fx
     *  The horizontal scaling factor.
     *
     * @param {Number} fy
     *  The vertical scaling factor.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.scale = function (fx, fy) {
        this._context.scale(fx, fy);
        return this;
    };

    /**
     * Changes the rotation angle of the coordinate system.
     *
     * @param {Number} a
     *  The rotation angle (radiant).
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.rotate = function (a) {
        // TODO: adjust global translation
        this._context.rotate(a);
        return this;
    };

    /**
     * Saves the current state of the canvas (colors, transformation, clipping,
     * etc.), invokes the passed callback function, and restores the canvas
     * state afterwards.
     *
     * @param {Function} callback
     *  The callback function to be invoked.
     *
     * @param {Object} [ctxt]
     *  The calling context for the callback function.
     *
     * @returns {Any}
     *  The return value of the callback function.
     */
    ContextWrapper.prototype.render = function (callback, ctxt) {
        this._context.save();
        var result = callback.call(ctxt);
        this._context.restore();
        return result;
    };

    /**
     * Saves the current state of the rendering context (all stroke, fill, and
     * text styles, transformation, clipping, etc.), sets up (or adds) a
     * rectangular clipping region, invokes the passed callback function, and
     * restores the saved context state afterwards.
     *
     * @param {Rectangle|Object|Null} rectangle
     *  The clipping rectangle. For convenience, can be set to null, if no
     *  clipping rectangle is available (conditional clipping without having to
     *  repeat the code executed in the callback).
     *
     * @param {Function} callback
     *  The callback function to be invoked with active clipping region.
     *
     * @param {Object} [ctxt]
     *  The calling context for the callback function.
     *
     * @returns {Any}
     *  The return value of the callback function.
     */
    ContextWrapper.prototype.clip = function (rectangle, callback, ctxt) {
        return this.render(function () {
            if (rectangle) {
                this._context.beginPath();
                this._context.rect(rectangle.left - this._x0, rectangle.top - this._y0, rectangle.width, rectangle.height);
                this._context.clip();
            }
            return callback.call(ctxt);
        }, this);
    };

    /**
     * Creates a new empty path object. Path objects are expected by other
     * rendering methods of this canvas wrapper. A caller may create as many
     * path objects as needed at the same time.
     *
     * @returns {Path}
     *  A new empty path object.
     */
    ContextWrapper.prototype.createPath = function () {
        return new Path();
    };

    /**
     * Renders the specified path into the wrapped canvas element. First, the
     * path will be filled according to the current fill style (see method
     * ContextWrapper.setFillStyle() for details). Afterwards, the path outline
     * will be rendered according to the current line style (see method
     * ContextWrapper.setLineStyle() for details).
     *
     * @param {Path} path
     *  The path to be rendered.
     *
     * @param {String} [mode='all']
     *  The rendering mode for the path. If specified, MUST be one of the
     *  following values:
     *  - 'all' (default): The path will be filled, and afterwards the path
     *      outline will be rendered.
     *  - 'fill': The path will be filled, but the path outline will NOT be
     *      rendered, regardless of the current line style.
     *  - 'stroke': The path outline will be rendered, but the path will NOT be
     *      filled, regardless of the current fill style.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.drawPath = function (path, mode) {

        // whether to actually fill and stroke the path
        var fill = mode !== 'stroke';
        var stroke = mode !== 'fill';

        // generate the path in the rendering context, and render the path
        if (fill || stroke) { this._pathGenerator.generatePath(path); }
        if (fill) { this._context.fill(); }
        if (stroke) { this._context.stroke(); }

        return this;
    };

    /**
     * Renders a straight line into the wrapped canvas element.
     *
     * @param {Number} x1
     *  The X coordinate of the starting point of the line.
     *
     * @param {Number} y1
     *  The Y coordinate of the starting point of the line.
     *
     * @param {Number} x2
     *  The X coordinate of the end point of the line.
     *
     * @param {Number} y2
     *  The Y coordinate of the end point of the line.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.drawLine = function (x1, y1, x2, y2) {
        return this.drawPath(new Path().pushLine(x1, y1, x2, y2), 'stroke');
    };

    /**
     * Renders a rectangle into the wrapped canvas element.
     *
     * @param {Number|Rectangle|Object} x
     *  The X coordinate of the starting point of the rectangle, or a complete
     *  rectangle with the numeric properties 'left', 'top', 'width', and
     *  'height'.
     *
     * @param {Number|String} y
     *  The Y coordinate of the starting point of the rectangle. If the
     *  parameter 'x' is a rectangle object, this parameter will be used as the
     *  optional parameter 'mode' (see below).
     *
     * @param {Number} w
     *  The width of the rectangle. If parameter 'x' is a rectangle object,
     *  this parameter will be ignored.
     *
     * @param {Number} h
     *  The height of the rectangle. If parameter 'x' is a rectangle object,
     *  this parameter will be ignored.
     *
     * @param {String} [mode='all']
     *  The rendering mode for the rectangle. See description of the method
     *  ContextWrapper.drawPath() for details.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.drawRect = function (x, y, w, h, mode) {
        if (typeof y === 'string') { mode = y; }
        return this.drawPath(new Path().pushRect(x, y, w, h), mode);
    };

    /**
     * Renders a circle into the wrapped canvas element.
     *
     * @param {Number} x
     *  The X coordinate of the center point of the circle.
     *
     * @param {Number} y
     *  The Y coordinate of the center point of the circle.
     *
     * @param {Number} r
     *  The radius of the circle.
     *
     * @param {String} [mode='all']
     *  The rendering mode for the circle. See description of the method
     *  ContextWrapper.drawPath() for details.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.drawCircle = function (x, y, r, mode) {
        return this.drawPath(new Path().pushCircle(x, y, r), mode);
    };

    /**
     * Renders an ellipse into the wrapped canvas element.
     *
     * @param {Number} x
     *  The X coordinate of the center point of the ellipse.
     *
     * @param {Number} y
     *  The Y coordinate of the center point of the ellipse.
     *
     * @param {Number} rx
     *  The radius of the ellipse on the X axis.
     *
     * @param {Number} ry
     *  The radius of the ellipse on the Y axis.
     *
     * @param {Number} [a=0]
     *  The rotation angle (radiant) of the ellipse.
     *
     * @param {String} [mode='all']
     *  The rendering mode for the circle. See description of the method
     *  ContextWrapper.drawPath() for details.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.drawEllipse = function (x, y, rx, ry, a, mode) {
        if (typeof a === 'string') { mode = a; a = 0; }
        return this.drawPath(new Path().pushEllipse(x, y, rx, ry, a), mode);
    };

    /**
     * Clears a rectangle (sets all pixels to full transparency) in the wrapped
     * canvas element.
     *
     * @param {Number|Rectangle|Object} x
     *  The X coordinate of the starting point of the rectangle, or a complete
     *  rectangle with the numeric properties 'left', 'top', 'width', and
     *  'height'.
     *
     * @param {Number} y
     *  The Y coordinate of the starting point of the rectangle (ignored, if
     *  parameter 'x' is a rectangle object).
     *
     * @param {Number} w
     *  The width of the rectangle (ignored, if parameter 'x' is a rectangle
     *  object).
     *
     * @param {Number} h
     *  The height of the rectangle (ignored, if parameter 'x' is a rectangle
     *  object).
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.clearRect = function (x, y, w, h) {
        if (isObject(x)) { y = x.top; w = x.width; h = x.height; x = x.left; }
        this._context.clearRect(x - this._x0, y - this._y0, w, h);
        return this;
    };

    /**
     * Draws the bitmap contents of the passed image source into the wrapped
     * canvas element.
     *
     * @param {CanvasImageSource|jQuery|Canvas} source
     *  The bitmap source object. Can be any DOM element supporting the DOM
     *  helper type CanvasImageSource (e.g. <img>, <canvas>, or <video>
     *  elements), or a jQuery collection wrapping a DOM element mentioned
     *  before, or an instance of the toolkit class Canvas itself.
     *
     * @param {Number|Object} p1
     *  Either the horizontal target coordinate as floating-point number, where
     *  to draw the entire unscaled source image into this context (in this
     *  case, parameter 'p2' MUST be a numeric coordinate too); or a rectangle
     *  object specifying the target area in this context to draw the scaled
     *  source image to.
     *
     * @param {Number|Object} [p2]
     *  If parameter 'p1' is a numeric coordinate, this parameter MUST BE the
     *  vertical target coordinate as floating-point number, where to draw the
     *  entire unscaled source image into this context. If parameter 'p1' is a
     *  rectangle, this parameter can be set to a rectangle specifying the
     *  partial area in the source image to be drawn into this context. If
     *  omitted, the entire source image will be scaled into the target
     *  rectangle. Note that in difference to the native method
     *  CanvasRenderingContext2D.drawImage(), the target rectangle (parameter
     *  'p1') ALWAYS precedes the source rectangle (parameter 'p2')! If the
     *  bitmap source object is an instance of the class Canvas, its global
     *  translation and resolution will be taken into account when interpreting
     *  this parameter.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.drawImage = function (source, p1, p2) {

        // pixel resolution in source canvas
        var sres = 1;
        // global translation in source canvas
        var sx0 = 0, sy0 = 0;

        // convert canvas wrapper or jQuery objects to canvas DOM element
        if (source instanceof Canvas) {
            sres = source.getResolution();
            var sourceRect = source.getRectangle();
            sx0 = sourceRect.left;
            sy0 = sourceRect.top;
            source = source.getNode()[0];
        } else if (source instanceof $) {
            source = source[0];
        }

        if (typeof p1 === 'number') {
            // coordinates as numbers
            this._context.drawImage(source, p1 - this._x0, p2 - this._y0);
        } else if (isObject(p2)) {
            // source and target rectangle (native method expects source before target)
            this._context.drawImage(source, (p2.left - sx0) * sres, (p2.top - sy0) * sres, p2.width * sres, p2.height * sres, p1.left - this._x0, p1.top - this._y0, p1.width, p1.height);
        } else {
            // target rectangle only
            this._context.drawImage(source, p1.left - this._x0, p1.top - this._y0, p1.width, p1.height);
        }

        return this;
    };

    /**
     * Renders the specified text into the wrapped canvas element.
     *
     * @param {String} text
     *  The text to be rendered.
     *
     * @param {Number} x
     *  The X coordinate of the starting point of the text.
     *
     * @param {Number} y
     *  The Y coordinate of the starting point of the text.
     *
     * @param {String} [mode='all']
     *  The rendering mode for the text. See description of the method
     *  ContextWrapper.drawPath() for details.
     *
     * @returns {ContextWrapper}
     *  A reference to this instance.
     */
    ContextWrapper.prototype.drawText = function (text, x, y, mode) {

        // fill the text, if a fill style has been set
        if (mode !== 'stroke') {
            this._context.fillText(text, x - this._x0, y - this._y0);
        }

        // stroke the text outline, if a stroke style has been set
        if (mode !== 'fill') {
            this._context.strokeText(text, x - this._x0, y - this._y0);
        }

        return this;
    };

    /**
     * Returns the width of the passed string.
     *
     * @param {String} text
     *  The text whose width will be calculated.
     *
     * @returns {Number}
     *  The width of the passed string, as floating-point pixels.
     */
    ContextWrapper.prototype.getTextWidth = function (text) {
        return this._context.measureText(text).width;
    };

    /**
     * Returns the width of the passed single character. To increase precision
     * for different browsers (some browsers return a rounded integral width
     * for single characters), the character will be repeated a hundred times,
     * and a hundredth of the width of the resulting string will be returned.
     *
     * @param {String} char
     *  A single character.
     *
     * @returns {Number}
     *  The width of the passed character, as floating-point pixels.
     */
    ContextWrapper.prototype.getCharacterWidth = function (char) {
        return this._context.measureText(Utils.repeatString(char, 100)).width / 100;
    };

    // class PixelWrapper =====================================================

    /**
     * This wrapper class provides a higher-level API for an ImageData object
     * of a canvas element.
     *
     * @constructor
     *
     * @param {CanvasRenderingContext2D} renderContext
     *  The wrapped rendering context of a canvas element.
     *
     * @param {Number} x0
     *  The global translation on the X axis that will be applied to all path
     *  operations.
     *
     * @param {Number} y0
     *  The global translation on the Y axis that will be applied to all path
     *  operations.
     */
    function PixelWrapper(renderContext, x0, y0) {

        this._context = renderContext;
        this._x0 = x0;
        this._y0 = y0;

        this._w = renderContext.canvas.width;
        this._h = renderContext.canvas.height;

        this._data = renderContext.getImageData(0, 0, this._w, this._h);
        this._color = TRANSPARENT;

    } // class PixelWrapper

    // private methods --------------------------------------------------------

    PixelWrapper.prototype._getOffset = function (x, y) {
        return 4 * ((y - this._y0) * this._w + x - this._x0);
    };

    PixelWrapper.prototype._writePixel = function (offset, color) {
        for (var i = 0; i < 4; offset += 1, i += 1) {
            this._data.data[offset] = color[i];
        }
    };

    PixelWrapper.prototype._writeRect = function (x, y, w, h, color) {
        if (isObject(x)) { y = x.top; w = x.width; h = x.height; x = x.left; }
        for (var y1 = y, y2 = y + h; y1 < y2; y1 += 1) {
            var offset = this._getOffset(x, y1);
            for (var i = 0; i < w; i += 1, offset += 4) {
                this._writePixel(offset, color);
            }
        }
    };

    // protected methods ------------------------------------------------------

    PixelWrapper.prototype._flush = function () {
        this._context.putImageData(this._data, 0, 0);
    };

    // public methods ---------------------------------------------------------

    /**
     * Returns the RGBA components of the specified canvas pixel.
     *
     * @param {Number} x
     *  The X coordinate of the pixel.
     *
     * @param {Number} y
     *  The Y coordinate of the pixel.
     *
     * @returns {Uint8ClampedArray}
     *  The RGBA components of the specified canvas pixel, as array with 4
     *  integers.
     */
    PixelWrapper.prototype.getPixelBytes = function (x, y) {
        var offset = this._getOffset(x, y);
        return this._data.slice(offset, offset + 4);
    };

    /**
     * Returns the CSS color of the specified canvas pixel.
     *
     * @param {Number} x
     *  The X coordinate of the pixel.
     *
     * @param {Number} y
     *  The Y coordinate of the pixel.
     *
     * @returns {String}
     *  The CSS color of the specified canvas pixel.
     */
    PixelWrapper.prototype.getPixelColor = function (x, y) {
        return RGBModel.parseBytes(this.getPixelBytes(x, y)).toCSS();
    };

    /**
     * Sets the color for new pixels drawn into the canvas.
     *
     * @param {String} color
     *  The new pixel color (any CSS color: as '#RRGGBB', '#RRGGBBAA', '#RGB',
     *  '#RGBA' style, or as 'rgb()', 'rgba()', 'hsl()', or 'hsla()' function
     *  style, or any predefined or system color name).
     *
     * @returns {PixelWrapper}
     *  A reference to this instance.
     */
    PixelWrapper.prototype.setColor = function (color) {
        var colorDesc = ColorUtils.parseColor(color);
        this._color = colorDesc ? colorDesc.bytes : TRANSPARENT;
    };

    /**
     * Sets the specified pixel to the current pixel color of this instance.
     *
     * @param {Number} x
     *  The X coordinate of the pixel.
     *
     * @param {Number} y
     *  The Y coordinate of the pixel.
     *
     * @returns {PixelWrapper}
     *  A reference to this instance.
     */
    PixelWrapper.prototype.drawPixel = function (x, y) {
        this._writePixel(this._getOffset(x, y), this._color);
        return this;
    };

    /**
     * Sets the specified pixel to full transparence.
     *
     * @param {Number} x
     *  The X coordinate of the pixel.
     *
     * @param {Number} y
     *  The Y coordinate of the pixel.
     *
     * @returns {PixelWrapper}
     *  A reference to this instance.
     */
    PixelWrapper.prototype.clearPixel = function (x, y) {
        this._writePixel(this._getOffset(x, y), TRANSPARENT);
        return this;
    };

    /**
     * Sets all pixels in the specified rectangle to the current pixel color of
     * this instance.
     *
     * @param {Number|Rectangle|Object} x
     *  The X coordinate of the starting point of the rectangle, or a complete
     *  rectangle with the numeric properties 'left', 'top', 'width', and
     *  'height'.
     *
     * @param {Number} y
     *  The Y coordinate of the starting point of the rectangle. If parameter
     *  'x' is a rectangle object, this parameter will be ignored.
     *
     * @param {Number} w
     *  The width of the rectangle. If parameter 'x' is a rectangle object,
     *  this parameter will be ignored.
     *
     * @param {Number} h
     *  The height of the rectangle. If parameter 'x' is a rectangle object,
     *  this parameter will be ignored.
     *
     * @returns {PixelWrapper}
     *  A reference to this instance.
     */
    PixelWrapper.prototype.fillRect = function (x, y, w, h) {
        this._writeRect(x, y, w, h, this._color);
        return this;
    };

    /**
     * Sets all pixels in the specified rectangle to full transparence.
     *
     * @param {Number|Rectangle|Object} x
     *  The X coordinate of the starting point of the rectangle, or a complete
     *  rectangle with the numeric properties 'left', 'top', 'width', and
     *  'height'.
     *
     * @param {Number} y
     *  The Y coordinate of the starting point of the rectangle. If parameter
     *  'x' is a rectangle object, this parameter will be ignored.
     *
     * @param {Number} w
     *  The width of the rectangle. If parameter 'x' is a rectangle object,
     *  this parameter will be ignored.
     *
     * @param {Number} h
     *  The height of the rectangle. If parameter 'x' is a rectangle object,
     *  this parameter will be ignored.
     *
     * @returns {PixelWrapper}
     *  A reference to this instance.
     */
    PixelWrapper.prototype.clearRect = function (x, y, w, h) {
        this._writeRect(x, y, w, h, TRANSPARENT);
        return this;
    };

    // class Canvas ===========================================================

    /**
     * A wrapper class for a canvas DOM element with additional convenience
     * methods to render into the canvas.
     *
     * @constructor
     *
     * @extends BaseObject
     *
     * @param {String|Object} windowId
     *  The identifier of the root window of the context application owning the
     *  instances of this class, or an object with a method 'getWindowId' that
     *  returns such a window identifier. Used for debugging and logging of
     *  running timers and pending promises in automated test environments.
     *
     * @param {Object} [initOptions]
     *  Optional parameters:
     *  @param {HTMLCanvasElement|jQuery} [initOptions.node]
     *      An existing <canvas> element to be wrapped by this instance. The
     *      current size of that canvas will be retained. By default, a new
     *      <canvas> element with zero size will be created automatically.
     *  @param {String} [initOptions.classes]
     *      Additional CSS classes (space-separated list) that will be added to
     *      the passed or created <canvas> element.
     */
    var Canvas = BaseObject.extend({ constructor: function (windowId, initOptions) {

        // self reference
        var self = this;

        // external canvas node wrapped by this instance
        var wrappedNode = Utils.getOption(initOptions, 'node', null);

        // the canvas DOM element, as jQuery object
        var $canvasNode = $(wrappedNode || '<canvas width="0" height="0">');

        // the canvas DOM element
        var canvasNode = $canvasNode[0];

        // the rendering context of the canvas
        var renderContext = canvasNode.getContext('2d');

        // additional classes for the canvas element
        var classes = Utils.getStringOption(initOptions, 'classes');

        // the current offset of the rectangle represented by the canvas element
        var x0 = 0;
        var y0 = 0;

        // the current size of the rectangle represented by the canvas element (bitmap size, not CSS size)
        var width = canvasNode.width;
        var height = canvasNode.height;

        // pixel resolution (physical pixels per canvas pixel in one direction)
        var res = 1;

        // base constructor ---------------------------------------------------

        BaseObject.call(this, windowId);

        // public methods -----------------------------------------------------

        /**
         * Returns the wrapped canvas element.
         *
         * @returns {jQuery}
         *  The wrapped canvas DOM element, as jQuery object.
         */
        this.getNode = function () {
            return $canvasNode;
        };

        /**
         * Returns the pixel resolution used by this canvas (the number of
         * physical pixels in each direction of the coordinate system
         * represented by one pixel offered by the API of this canvas).
         *
         * @returns {Number}
         *  The pixel resolution used by this canvas.
         */
        this.getResolution = function () {
            return res;
        };

        /**
         * Returns the rectangle currently represented by this canvas wrapper.
         *
         * @returns {Rectangle}
         *  The rectangle currently represented by this canvas wrapper.
         */
        this.getRectangle = function () {
            return new Rectangle(x0, y0, width, height);
        };

        /**
         * Returns a data URL with the current contents of the canvas.
         *
         * @returns {String}
         *  A data URL with the current contents of the canvas.
         */
        this.getDataURL = function () {
            return canvasNode.toDataURL();
        };

        /**
         * Changes the size and global translation of the canvas element, and
         * clears all its contents.
         *
         * @param {Rectangle|Object} rectangle
         *  The new rectangle represented by the canvas DOM element (with
         *  optional start position), with the following properties:
         *  @param {Number} [rectangle.left=0]
         *      Global horizontal translation, in pixels. A pixel drawn at this
         *      position will appear at the left border of the canvas area.
         *  @param {Number} [rectangle.top=0]
         *      Global vertical translation, in pixels. A pixel drawn at this
         *      position will appear at the top border of the canvas area.
         *  @param {Number} rectangle.width
         *      Width of the bitmap in the canvas area, in pixels. Will also be
         *      set as element size of the <canvas> DOM element.
         *  @param {Number} rectangle.height
         *      Height of the bitmap in the canvas area, in pixels. Will also
         *      be set as element size of the <canvas> DOM element.
         *
         * @param {Number} [resolution=1]
         *  Number of physical pixels in each direction of the coordinate
         *  system represented by one pixel offered by the API of this canvas.
         *
         * @returns {Canvas}
         *  A reference to this instance.
         */
        this.initialize = function (rectangle, resolution) {

            // rescue old settings for performance optimizations
            var oldCssWidth = $canvasNode.width();
            var oldCssHeight = $canvasNode.height();
            var oldRes = res;

            // set global translation
            x0 = Utils.getIntegerOption(rectangle, 'left', 0);
            y0 = Utils.getIntegerOption(rectangle, 'top', 0);
            width = rectangle.width;
            height = rectangle.height;

            // set new pixel resolution
            res = (typeof resolution === 'number') ? resolution : 1;

            // the new internal size of the canvas bitmap, according to pixel resolution
            var canvasWidth = Math.ceil(width * res);
            var canvasHeight = Math.ceil(height * res);
            var changedBitmapSize = (canvasNode.width !== canvasWidth) || (canvasNode.height !== canvasHeight);

            // do not touch canvas bitmap size, if it remains the same (may improve performance)
            if (changedBitmapSize) {
                canvasNode.width = canvasWidth;
                canvasNode.height = canvasHeight;
            }

            // initialize the visible size of the canvas element
            if ((oldCssWidth !== width) || (oldCssHeight !== height)) {
                $canvasNode.css({ width: width, height: height });
            }

            // initialize global transformation (has been reset after changing canvas size)
            if (changedBitmapSize || (oldRes !== res)) {
                try {
                    renderContext.setTransform(res, 0, 0, res, 0, 0);
                } catch (ex) {
                    Utils.error('Canvas.initialize(): context.setTransform failed (' + res + '): ' + ex.stack);
                }
            }

            // always clear all contents
            return this.clear();
        };

        /**
         * Changes the location covered by this canvas element. Moves the
         * bitmap contents of the canvas to the new location, if the current
         * location and the passed location overlap each other.
         *
         * @param {Rectangle|Object} rectangle
         *  The new location covered by this canvas element. See description
         *  of the method Canvas.initialize() for details.
         *
         * @param {Number} [scale=1]
         *  The scaling factor used when copying the old bitmap contents of
         *  this canvas element to the passed target location.
         *
         * @returns {Array<Rectangle>}
         *  The locations of all remaining canvas areas that could not be
         *  filled with the old contents of the canvas element.
         */
        this.relocate = function (rectangle, scale) {

            // convert parameter to an instance of Rectangle
            rectangle = Rectangle.from(rectangle);

            // effective scaling factor
            var scaleFactor = ((typeof scale === 'number') && (scale > 0)) ? scale : 1;
            // the old rectangle in the source canvas (previous zoom state)
            var sourceOldRectangle = this.getRectangle();
            // the new rectangle in the source canvas (scaled back to the previous zoom state)
            var sourceNewRectangle = rectangle.clone().scaleSelf(1 / scaleFactor, 1 / scaleFactor);
            // the common part of old and new rectangle in source canvas (previous zoom state)
            var sourceCommonRectangle = sourceOldRectangle.intersect(sourceNewRectangle);
            // the old rectangle in target canvas (scaled to the new zoom state)
            var targetOldRectangle = sourceOldRectangle.clone().scaleSelf(scaleFactor, scaleFactor);
            // the common part of old and new rectangle in target canvas (current zoom state)
            var targetCommonRectangle = targetOldRectangle.intersect(rectangle);

            // if possible, copy common areas to secondary canvas, and swap the canvases
            if (isValidRectangle(sourceCommonRectangle) && isValidRectangle(targetCommonRectangle)) {

                // copy bitmap data to temporary canvas (using a temporary canvas element
                // is up to 100 times faster than using getImageData/putImageData)
                relocateCanvas.initialize(targetCommonRectangle, res).render(function (relocContext) {
                    relocContext.drawImage(self, targetCommonRectangle, sourceCommonRectangle);
                });

                // initialize this canvas element, and copy the (already scaled) image data back
                this.initialize(rectangle, res).render(function (thisContext) {
                    thisContext.drawImage(relocateCanvas, targetCommonRectangle);
                });

                // return the remaining empty areas in the canvas
                return rectangle.remaining(targetOldRectangle);
            }

            // nothing to copy (no common areas)
            this.initialize(rectangle);

            // return the entire area as resulting empty space
            return [rectangle];
        };

        /**
         * Clears the contents of the entire canvas, i.e. fills the canvas with
         * fully transparent pixels.
         *
         * @returns {Canvas}
         *  A reference to this instance.
         */
        this.clear = function () {
            renderContext.clearRect(0, 0, width, height);
            return this;
        };

        /**
         * Saves the current state of the canvas (colors, transformation,
         * clipping, etc.), invokes the passed callback function, and restores
         * the canvas state afterwards.
         *
         * @param {Rectangle|Object} [clipRect]
         *  An optional clipping rectangle. Can be omitted completely (instead
         *  of being set to null or undefined).
         *
         * @param {Function} callback
         *  The callback function to be invoked. Receives the following
         *  parameters:
         *  (1) {ContextWrapper} context
         *      The context wrapper of the canvas element, which provides a
         *      more convenient rendering API than the native rendering context
         *      of the canvas, and that adds polyfills for missing features of
         *      the current browser.
         *  (2) {Number} width
         *      The current width of the canvas area, in pixels.
         *  (3) {Number} height
         *      The current height of the canvas area, in pixels.
         *
         * @param {Object} [ctxt]
         *  The calling context for the callback function.
         *
         * @returns {Any}
         *  The return value of the callback function.
         */
        this.render = function (clipRect, callback, ctxt) {

            // Do not set the global translation at the native canvas element for large offsets, to workaround
            // the limited floating-point precision provided by the canvas. Browsers usually only implement
            // single-precision instead of double-precision as required by the W3C (see bug 35653, and
            // http://stackoverflow.com/a/8874802).
            var translateManually = (Math.abs(x0) > 4e6) || (Math.abs(y0) > 4e6);
            var globalX0 = translateManually ? x0 : 0;
            var globalY0 = translateManually ? y0 : 0;
            if (!translateManually) { renderContext.translate(-x0, -y0); }

            // the context wrapper for this invocation
            var contextWrapper = new ContextWrapper(renderContext, globalX0, globalY0);

            // adjust arguments (missing clipping rectangle)
            if (isFunction(clipRect)) {
                ctxt = callback;
                callback = clipRect;
                clipRect = null;
            }

            // apply clipping if specified, invoke callback function
            try {
                return contextWrapper.clip(clipRect, function () {
                    return callback.call(ctxt, contextWrapper, width, height);
                });
            } finally {
                if (!translateManually) { renderContext.translate(x0, y0); }
            }
        };

        /**
         * Invokes the passed callback function, and provides a helper context
         * needed to manipulate the single pixels of this canvas.
         *
         * @param {Function} callback
         *  The callback function to be invoked. Receives the following
         *  parameters:
         *  (1) {PixelWrapper} context
         *      The pixel data wrapper of the canvas element, which provides a
         *      more convenient rendering API than the native ImageData class.
         *  (2) {Number} width
         *      The current width of the canvas area, in pixels.
         *  (3) {Number} height
         *      The current height of the canvas area, in pixels.
         *
         * @param {Object} [ctxt]
         *  The calling context for the callback function.
         *
         * @returns {Any}
         *  The return value of the callback function.
         */
        this.renderPixels = function (callback, ctxt) {

            // the image data wrapper for this invocation
            var pixelWrapper = new PixelWrapper(renderContext, x0, y0);

            // invoke callback function, and flush the pixel data
            try {
                return callback.call(ctxt, pixelWrapper, width, height);
            } finally {
                pixelWrapper._flush();
            }
        };

        // initialization -----------------------------------------------------

        // add CSS class names passed to the constructor
        if (classes) { $canvasNode.addClass(classes); }

        // add polyfills for missing native CanvasRenderingContext2D methods
        if (!isFunction(renderContext.ellipse)) {
            renderContext.ellipse = function (x, y, rx, ry, a0, a1, a2, ccw) {
                this.translate(x, y);
                this.rotate(a0);
                this.scale(rx, ry);
                this.arc(0, 0, 1, a1, a2, ccw);
                this.scale(1 / rx, 1 / ry);
                this.rotate(-a0);
                this.translate(-x, -y);
            };
        }

        // destroy all class members on destruction
        this.registerDestructor(function () {
            if (!wrappedNode) { $canvasNode.remove(); }
            self = initOptions = wrappedNode = $canvasNode = canvasNode = renderContext = null;
        });

    } }); // class Canvas

    // global initialization ==================================================

    // create a global canvas object for relocation
    relocateCanvas = new Canvas(BaseObject.SINGLETON);

    // exports ================================================================

    return Canvas;

});
